# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from .test_module import add_currency_rate, create_currency

__all__ = ['create_currency', 'add_currency_rate']
